# -*- coding: utf-8 -*-

"""
Modified from werkzeug.debug
"""

# -- stdlib --
from urllib.parse import urljoin
import mimetypes
import pkgutil
import sys
import types

# -- third party --
from flask import Blueprint, Response, current_app, request

# -- own --
from ..utils.check import check_type
from .tbtools import Traceback


# -- code --
blueprint = Blueprint("openbayes-serv-debugger", __name__)


@blueprint.route('/.debug/console/<tb>', methods=['GET'])
def debug_page(tb):
    return pkgutil.get_data(__name__, 'static/index.html')


@blueprint.route('/.debug/static/<path:filename>', methods=['GET'])
def static_files(filename):
    typ, enc = mimetypes.guess_type(filename)
    if typ and enc:
        mime = f'{typ}+{enc}'
    elif typ and not enc:
        mime = typ
    else:
        mime = 'application/octet-stream'

    return Response(
        pkgutil.get_data(__name__, f'static/{filename}'),
        mimetype=mime,
    )


@blueprint.route('/.debug/traceback/<tb>', methods=['GET'])
def traceback(tb):
    state = current_app.config['OPENBAYES_SERVING_DEBUGGER_STATE']
    if tb not in state.tracebacks:
        return {}, 404

    return state.tracebacks[tb].render()

    return tb


@blueprint.route('/.debug/frame/<frame>/exec', methods=['POST'])
def frame_exec(frame):
    state = current_app.config['OPENBAYES_SERVING_DEBUGGER_STATE']
    if frame not in state.frames:
        return {}, 404

    err = check_type({'code': str, ...: '!'}, request.json)
    if err:
        return {'error': 'Invalid input'}, 400

    frame = state.frames[frame]
    return frame.console.eval(request.json['code'])


class ManualCapture(Exception):
    pass


class DebuggerState(object):

    def __init__(self):
        self.frames = {}
        self.tracebacks = {}

    def collect_exception(self):
        exc_type, exc_value, tb = sys.exc_info()
        tb = Traceback(exc_type, exc_value, tb)
        tb.filter_hidden_frames()
        return self._save(tb)

    def collect_current(self):
        from .tbtools import Traceback
        tb = self._frame2tb(sys._getframe(1))
        e = ManualCapture("手动捕获的执行轨迹")
        tb = Traceback(type(e), e, tb)
        tb.filter_hidden_frames()
        return self._save(tb)

    def _frame2tb(self, f):
        flatted = []
        while f:
            flatted.append(f)
            f = f.f_back
        tb = None
        for f in flatted:
            tb = types.TracebackType(tb, f, f.f_lasti, f.f_lineno)

        return tb

    def _save(self, tb):
        for frame in tb.frames:
            self.frames[str(frame.id)] = frame

        tag = str(tb.id)
        self.tracebacks[tag] = tb

        root = current_app.config.get('OPENBAYES_EXTERNAL_ROOT') or request.url_root

        return {
            'id': tag,
            'url': urljoin(f'{root}/', f'.debug/console/{tag}'),
        }


def install(app, url_prefix=None):
    app.config['OPENBAYES_SERVING_DEBUGGER_STATE'] = DebuggerState()
    app.register_blueprint(blueprint, url_prefix=url_prefix)



# -*- coding: utf-8 -*-

# -- stdlib --
import datetime
import logging
import os
import sys

# -- third party --
# -- own --
from .escapes import escape_codes


# -- code --
class ColorLogFormatter(logging.Formatter):
    def __init__(self, use_color=True):
        super().__init__()
        self.use_color = use_color
        self.color_mapping = {
            'CRITICAL': 'bold_red',
            'ERROR': 'red',
            'WARNING': 'yellow',
            'INFO': 'green',
            'DEBUG': 'blue',
        }

    def format(self, rec):

        if rec.exc_info:
            s = []
            s.append('>>>>>>' + '-' * 74)
            s.append(self._format(rec))
            import traceback
            s.append(''.join(traceback.format_exception(*rec.exc_info)).strip())
            s.append('<<<<<<' + '-' * 74)
            return '\n'.join(s)
        else:
            return self._format(rec)

    def _format(self, rec):
        import time

        rec.message = rec.getMessage()
        lvl = rec.levelname
        prefix = '[{} {} {}:{}]'.format(
            lvl[0],
            time.strftime('%y%m%d %H:%M:%S'),
            rec.module,
            rec.lineno,
        )
        if self.use_color:
            E = escape_codes
            M = self.color_mapping
            prefix = f"{E[M[lvl]]}{prefix}{E['reset']}"

        return f'{prefix} {rec.message}'


def patch_formatter(logger):
    use_color = sys.platform == 'win32' or os.isatty(1)
    for h in logger.handlers:
        formatter = ColorLogFormatter(use_color=use_color)
        h.setFormatter(formatter)


def init(level):
    root = logging.getLogger()
    logging.DEBUG
    root.setLevel(0)

    hdlr = logging.StreamHandler(sys.stdout)
    hdlr.setLevel(level)
    root.addHandler(hdlr)

    patch_formatter(root)

    logging.getLogger('sentry.errors').setLevel(1000)

    use_color = sys.platform == 'win32' or os.isatty(1)
    formatter = ColorLogFormatter(use_color=use_color)
    hdlr.setFormatter(formatter)



# -*- coding: utf-8 -*-

# -- stdlib --
import types

# -- third party --
# -- own --

# -- code --


class CheckTypeFailed(Exception):
    def __init__(self):
        Exception.__init__(self)
        self.path = []

    def path_string(self):
        s = ['.']
        for v in self.path:
            if isinstance(v, int):
                s.append(f'[{v}]')
            elif isinstance(v, str):
                if s[-1] == '.':
                    s.append(v)
                else:
                    s.append(f'.{v}')

        return ''.join(s)

    def finalize(self):
        self.message = self.path_string()
        self.args = (self.message, )


def _check(cond):
    if not cond:
        raise CheckTypeFailed


def _check_isinstance(obj, cls):
    try:
        _check(isinstance(obj, cls))
    except TypeError as e:
        raise CheckTypeFailed from e


_check_key_not_exists = object()


def check_type_exc(pattern, obj, path=None):
    try:
        if isinstance(pattern, (list, tuple)):
            _check_isinstance(obj, (list, tuple))
            if len(pattern) == 2 and pattern[-1] is ...:
                cls = pattern[0]
                for i, v in enumerate(obj):
                    check_type_exc(cls, v, i)
            else:
                _check(len(pattern) == len(obj))
                for i, (cls, v) in enumerate(zip(pattern, obj)):
                    check_type_exc(cls, v, i)

        elif isinstance(pattern, dict):
            _check_isinstance(obj, dict)
            if ... in pattern:
                pattern = dict(pattern)
                match = pattern.pop(...)
            else:
                match = '!'

            if match in set('?!='):
                lkeys = set(pattern.keys())
                rkeys = set(obj.keys())

                if match == '!':
                    iterkeys = lkeys
                elif match == '?':
                    iterkeys = lkeys & rkeys
                elif match == '=':
                    _check(lkeys == rkeys)
                    iterkeys = lkeys
                else:
                    assert False, 'WTF?!'

                for k in iterkeys:
                    check_type_exc(pattern[k], obj.get(k, _check_key_not_exists), k)

            elif match is ...:
                assert len(pattern) == 1, 'Invalid dict pattern'
                kt, vt = list(pattern.items())[0]
                for k in obj:
                    check_type_exc(kt, k, '<%s>' % kt.__name__)
                    check_type_exc(vt, obj[k], k)
            else:
                assert False, 'Invalid dict match type'

        else:
            if issubclass(type(pattern), types.FunctionType):
                try:
                    _check(pattern(obj))
                except Exception as e:
                    raise CheckTypeFailed from e
            elif issubclass(type(pattern), (int, str, bytes, tuple)):
                _check(obj == pattern)
            else:
                _check_isinstance(obj, pattern)

    except CheckTypeFailed as e:
        if path is not None:
            e.path.insert(0, path)
        else:
            e.finalize()

        raise


def check_type(pattern, obj):
    try:
        check_type_exc(pattern, obj)
        return None
    except CheckTypeFailed as e:
        return e.path_string()



# -*- coding: utf-8 -*-

"""
Modified from werkzeug.debug.tbtools
"""

# -- stdlib --
import codecs
import inspect
import os
import re
import sys
import sysconfig
import traceback

# -- third party --
from werkzeug._compat import range_type, reraise, string_types, text_type, to_native, to_unicode
from werkzeug.filesystem import get_filesystem_encoding
from werkzeug.utils import cached_property

# -- own --
from .console import Console

# -- code --
_coding_re = re.compile(br"coding[:=]\s*([-\w.]+)")
_line_re = re.compile(br"^(.*?)$", re.MULTILINE)
_funcdef_re = re.compile(r"^(\s*def\s)|(.*(?<!\w)lambda(:|\s))|^(\s*@)")
UTF8_COOKIE = b"\xef\xbb\xbf"

system_exceptions = (SystemExit, KeyboardInterrupt)
try:
    system_exceptions += (GeneratorExit,)
except NameError:
    pass


class Traceback(object):
    """Wraps a traceback."""

    def __init__(self, exc_type, exc_value, tb):
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.tb = tb

        exception_type = exc_type.__name__
        if exc_type.__module__ not in {"builtins", "__builtin__", "exceptions"}:
            exception_type = exc_type.__module__ + "." + exception_type
        self.exception_type = exception_type

        self.groups = []
        memo = set()
        while True:
            self.groups.append(Group(exc_type, exc_value, tb))
            memo.add(id(exc_value))
            exc_value = exc_value.__cause__ or exc_value.__context__
            if exc_value is None or id(exc_value) in memo:
                break
            exc_type = type(exc_value)
            tb = exc_value.__traceback__
        self.groups.reverse()
        self.frames = [frame for group in self.groups for frame in group.frames]

    def filter_hidden_frames(self):
        """Remove the frames according to the paste spec."""
        for group in self.groups:
            group.filter_hidden_frames()

        self.frames[:] = [frame for group in self.groups for frame in group.frames]

    @property
    def exception(self):
        """String representation of the final exception."""
        return self.groups[-1].exception

    def log(self, logfile=None):
        """Log the ASCII traceback into a file object."""
        if logfile is None:
            logfile = sys.stderr
        tb = self.plaintext.rstrip() + u"\n"
        logfile.write(to_native(tb, "utf-8", "replace"))

    def render_groups(self, include_title=True):
        """Render the traceback for the interactive console."""
        if not self.frames:
            frames = []
        else:
            frames = [group.render() for group in self.groups]

        return frames

    def render(self):
        return {
            "exception": self.exception,
            "exception_type": self.exception_type,
            "groups": self.render_groups(),
            "plaintext": self.plaintext,
            "id": str(self.id),
        }

    @cached_property
    def plaintext(self):
        import traceback
        return ''.join(traceback.format_exception(self.exc_type, self.exc_value, self.tb))

    @property
    def id(self):
        return id(self)


class Group(object):
    """A group of frames for an exception in a traceback. On Python 3,
    if the exception has a ``__cause__`` or ``__context__``, there are
    multiple exception groups.
    """

    def __init__(self, exc_type, exc_value, tb):
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.has_cause = exc_value.__cause__ is not None
        self.has_context = exc_value.__context__ is not None
        self.frames = []

        while tb is not None:
            self.frames.append(Frame(exc_type, exc_value, tb))
            tb = tb.tb_next

    def filter_hidden_frames(self):
        new_frames = []
        hidden = False

        for frame in self.frames:
            hide = frame.hide
            if hide in ("before", "before_and_this"):
                new_frames = []
                hidden = False
                if hide == "before_and_this":
                    continue
            elif hide in ("reset", "reset_and_this"):
                hidden = False
                if hide == "reset_and_this":
                    continue
            elif hide in ("after", "after_and_this"):
                hidden = True
                if hide == "after_and_this":
                    continue
            elif hide or hidden:
                continue
            new_frames.append(frame)

        # if we only have one frame and that frame is from the codeop
        # module, remove it.
        if len(new_frames) == 1 and self.frames[0].module == "codeop":
            del self.frames[:]

        # if the last frame is missing something went terrible wrong :(
        elif self.frames[-1] in new_frames:
            self.frames[:] = new_frames

    @property
    def exception(self):
        """String representation of the exception."""
        buf = traceback.format_exception_only(self.exc_type, self.exc_value)
        rv = "".join(buf).strip()
        return to_unicode(rv, "utf-8", "replace")

    def render(self):
        relation = 'root'
        if self.has_cause:
            relation = 'cause'
        elif self.has_context:
            relation = 'context'

        return {
            'relation': relation,
            'exception': self.exception,
            'frames': [frame.render() for frame in self.frames],
        }

    def render_text(self):
        out = []
        out.append(u"Traceback (most recent call last):")
        for frame in self.frames:
            out.append(frame.render_text())
        out.append(self.exception)
        return u"\n".join(out)


class Frame(object):
    """A single frame in a traceback."""

    def __init__(self, exc_type, exc_value, tb):
        self.lineno = tb.tb_lineno
        self.function_name = tb.tb_frame.f_code.co_name
        self.locals = tb.tb_frame.f_locals
        self.globals = tb.tb_frame.f_globals

        try:
            zelf = self.locals['self']
            class_name = zelf.__class__.__name__
            self.function_name = f'{class_name}.{self.function_name}'
        except Exception:
            pass

        fn = inspect.getsourcefile(tb) or inspect.getfile(tb)
        if fn[-4:] in (".pyo", ".pyc"):
            fn = fn[:-1]
        # if it's a file on the file system resolve the real filename.
        if os.path.isfile(fn):
            realpath = os.path.realpath(fn)
        else:
            realpath = fn

        self.filename = to_unicode(fn, get_filesystem_encoding())
        self.realpath = to_unicode(realpath, get_filesystem_encoding())
        self.abbrname = self.abbr_filename(self.realpath) or self.filename
        self.module = self.globals.get("__name__", self.locals.get("__name__"))
        self.loader = self.globals.get("__loader__", self.locals.get("__loader__"))
        self.code = tb.tb_frame.f_code

        # support for paste's traceback extensions
        self.hide = self.locals.get("__traceback_hide__", False)

    def render(self):
        """Render a single frame in a traceback."""
        return {
            "id": str(self.id),
            "filename": self.abbrname,
            "realpath": self.realpath,
            "lineno": self.lineno,
            "function": self.function_name,
            "context": self.render_line_context(),
            "is_library": self.is_library,
        }

    @cached_property
    def is_library(self):
        return any(
            self.filename.startswith(path) for path in sysconfig.get_paths().values()
        )

    def abbr_filename(self, fn):
        cwd = os.getcwd()
        if fn.startswith(os.getcwd()):
            return fn[len(cwd)+1:]

        if '/site-packages/' in fn:
            fn = fn.split('/site-packages/')[-1]
            fn = self._decorate_lib(fn)
            return fn

        for prefix in sysconfig.get_paths().values():
            if fn.startswith(prefix):
                return self._decorate_lib(fn[len(prefix)+1:])

        return None

    def _decorate_lib(self, fn):
        parts = fn.split('/', 1)
        if len(parts) > 1:
            fn = f'[{parts[0]}]:{parts[1]}'
        else:
            fn = f'[{parts[0]}]'

        return fn

    def render_text(self):
        return u'  File "%s", line %s, in %s\n    %s' % (
            self.filename,
            self.lineno,
            self.function_name,
            self.current_line.strip(),
        )

    def render_line_context(self):
        lines = []

        context = 5
        l, r = (self.lineno - 1) - context, (self.lineno - 1) + context + 1
        l = max(0, l)
        r = min(len(self.sourcelines), r)

        for n, line in zip(range(l, r), self.sourcelines[l:r]):
            lines.append({
                'lineno': n + 1,
                'line': line,
            })

        return lines

    def eval(self, code, mode="single"):
        """Evaluate code in the context of the frame."""
        if isinstance(code, string_types):
            if isinstance(code, text_type):  # noqa
                code = UTF8_COOKIE + code.encode("utf-8")
            code = compile(code, "<interactive>", mode)
        return eval(code, self.globals, self.locals)

    @cached_property
    def sourcelines(self):
        """The sourcecode of the file as list of unicode strings."""
        # get sourcecode from loader or file
        source = None
        if self.loader is not None:
            try:
                if hasattr(self.loader, "get_source"):
                    source = self.loader.get_source(self.module)
                elif hasattr(self.loader, "get_source_by_code"):
                    source = self.loader.get_source_by_code(self.code)
            except Exception:
                # we munch the exception so that we don't cause troubles
                # if the loader is broken.
                pass

        if source is None:
            try:
                with open(
                    to_native(self.filename, get_filesystem_encoding()), mode="rb"
                ) as f:
                    source = f.read()
            except IOError:
                return []

        # already unicode?  return right away
        if isinstance(source, text_type):
            return source.splitlines()

        # yes. it should be ascii, but we don't want to reject too many
        # characters in the debugger if something breaks
        charset = "utf-8"
        if source.startswith(UTF8_COOKIE):
            source = source[3:]
        else:
            for idx, match in enumerate(_line_re.finditer(source)):
                match = _coding_re.search(match.group())
                if match is not None:
                    charset = match.group(1)
                    break
                if idx > 1:
                    break

        # on broken cookies we fall back to utf-8 too
        charset = to_native(charset)
        try:
            codecs.lookup(charset)
        except LookupError:
            charset = "utf-8"

        return source.decode(charset, "replace").splitlines()

    @cached_property
    def console(self):
        return Console(self.globals, self.locals)

    @property
    def current_line(self):
        try:
            return self.sourcelines[self.lineno - 1]
        except IndexError:
            return u""

    @property
    def id(self):
        return id(self)



# -*- coding: utf-8 -*-
"""
    Source: http://docs.python.org/3/library/datetime.html → "Example tzinfo classes"
    Idea:   http://stackoverflow.com/a/2071364/183995
"""


# -- stdlib --
from datetime import datetime as dt, timedelta, tzinfo
import logging
import time

# -- third party --
# -- own --

# -- code --
ZERO = timedelta(0)

STDOFFSET = timedelta(seconds=-time.timezone)
if time.daylight:
    DSTOFFSET = timedelta(seconds=-time.altzone)
else:
    DSTOFFSET = STDOFFSET

DSTDIFF = DSTOFFSET - STDOFFSET


class LocalTimezone(tzinfo):

    def utcoffset(self, dt):
        if self._isdst(dt):
            return DSTOFFSET
        else:
            return STDOFFSET

    def dst(self, dt):
        if self._isdst(dt):
            return DSTDIFF
        else:
            return ZERO

    def tzname(self, dt):
        return time.tzname[self._isdst(dt)]

    def _isdst(self, dt):
        tt = (dt.year, dt.month, dt.day,
              dt.hour, dt.minute, dt.second,
              dt.weekday(), 0, 0)
        stamp = time.mktime(tt)
        tt = time.localtime(stamp)
        return tt.tm_isdst > 0


Local = LocalTimezone()

# ------------------------------------------------------------------------------

"""
Apache-like combined logging for WSGI Web Applications.

Homepage: https://github.com/pklaus/wsgi-request-logger

Copyright (c) 2013, Philipp Klaus. All rights reserved.
Copyright (c) 2007-2011 L. C. Rees. All rights reserved.

License: BSD (see LICENSE for details)
"""

__all__ = ['WSGILogger', 'ApacheFormatter']

try:
    clock = time.perf_counter
except AttributeError:
    clock = time.time


def ApacheFormatter(with_response_time=True):
    """ A factory that returns the wanted formatter """
    if with_response_time:
        return ApacheFormatters.format_with_response_time
    else:
        return ApacheFormatters.format_NCSA_log


class ApacheFormatters(object):
    @staticmethod
    def format_NCSA_log(status_code, environ, content_length, **kwargs):
        """
          Apache log format 'NCSA extended/combined log':
          "%h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-agent}i\""
          see http://httpd.apache.org/docs/current/mod/mod_log_config.html#formats
        """

        # Let's collect log values
        val = dict()
        ip_header = kwargs.get('ip_header', None)
        if ip_header:
            val['host'] = environ.get('HTTP_X_FORWARDED_FOR', '')
        else:
            val['host'] = environ.get('REMOTE_ADDR', '')
        val['logname'] = '-'
        val['user'] = '-'
        date = dt.now(tz=Local)
        month = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
                 "Aug", "Sep", "Oct", "Nov", "Dec"][date.month - 1]
        val['time'] = date.strftime("%d/{0}/%Y:%H:%M:%S %z".format(month))
        val['request'] = "{0} {1} {2}".format(
            environ.get('REQUEST_METHOD', ''),
            environ.get('PATH_INFO', ''),
            environ.get('SERVER_PROTOCOL', '')
        )
        val['status'] = status_code
        val['size'] = content_length
        val['referer'] = environ.get('HTTP_REFERER', '')
        val['agent'] = environ.get('HTTP_USER_AGENT', '')

        # see http://docs.python.org/3/library/string.html#format-string-syntax
        FORMAT = '{host} {logname} {user} [{time}] "{request}" '
        FORMAT += '{status} {size} "{referer}" "{agent}"'
        return FORMAT.format(**val)

    @staticmethod
    def format_with_response_time(*args, **kwargs):
        """
          The dict kwargs should contain 'rt_us', the response time in milliseconds.
          This is the format for TinyLogAnalyzer:
          https://pypi.python.org/pypi/TinyLogAnalyzer
        """
        rt_us = kwargs.get('rt_us')
        return ApacheFormatters.format_NCSA_log(*args, **kwargs) + " {0}/{1}".format(int(rt_us/1000000), rt_us % 1000000)


class WSGILogger(object):
    """ This is the generalized WSGI middleware for any style request logging. """

    def __init__(self, application, formatter=ApacheFormatter(), propagate=True, ip_header=None, **kwargs):
        self.logger = logging.getLogger('requestlogger')
        self.formatter = formatter
        self.ip_header = ip_header
        self.application = application

    def __call__(self, environ, start_response):
        start = clock()
        status_codes = []
        content_lengths = []

        def custom_start_response(status, response_headers, exc_info=None):
            status_codes.append(int(status.partition(' ')[0]))
            for name, value in response_headers:
                if name.lower() == 'content-length':
                    content_lengths.append(int(value))
                    break
            return start_response(status, response_headers, exc_info)
        retval = self.application(environ, custom_start_response)
        runtime = int((clock() - start) * 10**6)
        content_length = content_lengths[0] if content_lengths else len(
            b''.join(retval))
        msg = self.formatter(
            status_codes[0], environ, content_length, ip_header=self.ip_header, rt_us=runtime)
        self.logger.info(msg)
        return retval



# -*- coding: utf-8 -*-

# -- stdlib --
from urllib.parse import urlparse
import argparse
import inspect
import json
import logging
import os
import re
import sys

# -- third party --
from flask import Blueprint, Flask, Response, current_app, jsonify, request
from werkzeug.exceptions import HTTPException
import msgpack
import requests
import uvicorn

# -- own --
from . import debug
from .common import detect_env
from .utils.log import init as init_logging, patch_formatter as patch_log_formatter


# -- code --
log = logging.getLogger('openbayes_serving.serv')


class ServingError(Exception):

    def __init__(self, message, status_code=400, payload={}):
        Exception.__init__(self)
        self.message = message
        self.status_code = status_code
        self.payload = {**payload, 'error': message}


# -----------------------------
class ONNXInit:
    @staticmethod
    def register_parameters(parser):
        env = detect_env()
        if env == 'local':
            path = os.getcwd()
        elif env == 'gear':
            path = os.getcwd()
        elif env == 'production':
            path = '/mnt/project'
        else:
            raise Exception('WTF!')

        parser.add_argument(
            '--model-path',
            default=path,
            help='ONNX 模型（.onnx 文件）搜索地址',
        )

    @staticmethod
    def init():
        search_path = current_app.config['OPENBAYES_SERVING_RUN_OPTIONS'].model_path
        log.debug(f'ONNX: 在 {search_path} 搜寻 ONNX 模型...')

        path = None
        if os.path.isdir(search_path):
            for fn in os.listdir(search_path):
                if not fn.endswith('.onnx'):
                    continue
                path = os.path.join(search_path, fn)
                break
            else:
                raise Exception('没有找到 ONNX 模型')
        elif os.path.isfile(search_path):
            path = search_path
        else:
            raise Exception('没有找到 ONNX 模型')

        log.info('正在加载 ONNX 模型：%s ...', path)

        try:
            import onnxruntime
        except ImportError:
            log.exception('无法加载 onnxruntime 库，请确认已经安装了它。如果没有使用特定的包管理工具，可以通过 `pip install onnxruntime` 安装。')
            sys.exit(1)

        # TODO: plain onnxruntime object or a wrapped one?
        model = onnxruntime.InferenceSession(path)
        return model


class ConfigInit:
    @staticmethod
    def register_parameters(parser):
        parser
        pass

    @staticmethod
    def init():
        return Configurator()

INIT_ARG_HANDLERS = {
    'onnx': ONNXInit,
    'config': ConfigInit,
}


def parse_type(tp):
    if not tp or tp == '*/*':
        return 'octet-stream'

    m = re.match(r'^application/([a-z0-9\.\-]+)', tp)
    if not m:
        return 'octet-stream'

    enc, *_ = m.groups()

    COALESCE = {
        'octet-stream':    'octet-stream',
        'msgpack':         'msgpack',
        'x-msgpack':       'msgpack',
        'vnd.msgpack':     'msgpack',
        'vnd.messagepack': 'msgpack',
        'json':            'json',
    }

    return COALESCE.get(enc, 'octet-stream')


def get_payload():
    tp = parse_type(request.mimetype)

    if not tp:
        raise ServingError("无效的 Content-Type", 415)

    if tp == 'msgpack':
        try:
            payload = msgpack.unpackb(request.get_data(), raw=False)
        except Exception as e:
            raise ServingError("无法解析输入的 MessagePack 消息", 400) from e
    elif tp == 'json':
        try:
            payload = json.loads(request.get_data())
        except Exception as e:
            raise ServingError("无法解析输入的 JSON 消息", 400) from e
    elif tp == 'octet-stream':
        payload = request.get_data()
    else:
        raise Exception(f'BUG: unknown mime type {tp}')

    return tp, payload


def get_json():
    tp, payload = get_payload()
    if tp not in ('msgpack', 'json'):
        raise ServingError('输入必须是 JSON 或者 MessagePack 格式')

    return payload


PREDICT_ARG_HANDLERS = {
    'data': lambda: request.data,
    'json': get_json,
    'payload': lambda: get_payload()[1],
    'params': lambda: request.args,
    'headers': lambda: request.headers,
    'request': lambda: request,
}

predict_blueprint = Blueprint('openbayes-serv-predict', __name__)


CORS_HEADERS = {
    'Access-Control-Allow-Origin': '*',
    'Access-Control-Allow-Methods': 'GET,POST,PUT,DELETE,OPTIONS',
    'Access-Control-Allow-Credentials': 'true',
    'Access-Control-Allow-Headers': 'Authorization,Accept,Content-Type,Accept-Encoding,User-Agent',
    'Access-Control-Max-Age': '86400',
}


@predict_blueprint.route('', methods=['GET'], provide_automatic_options=False)
def help_message():
    return {'message': 'Serving 服务正常工作，可以发送 POST 请求到此地址进行推理。'}, 200, _cors_headers()


@predict_blueprint.route('', methods=['OPTIONS'], provide_automatic_options=False)
def request_options():
    return b'', 200, _cors_headers()


def _cors_headers():
    headers = CORS_HEADERS
    origin = request.headers.get('Origin')
    if origin:
        headers = {**CORS_HEADERS, 'Access-Control-Allow-Origin': origin}
    return headers


@predict_blueprint.route('', methods=['POST'], provide_automatic_options=False)
def model_predict():
    try:
        obj = current_app.config['OPENBAYES_SERVING_PREDICTOR_OBJECT']
        args = current_app.config['OPENBAYES_SERVING_PREDICT_ARGS']
        kwargs = {k: PREDICT_ARG_HANDLERS[k]() for k in args}
        # TODO: Parse Accept and optionally send msgpack payload
        rst = obj.predict(**kwargs)
        if isinstance(rst, (dict, list, tuple)):
            resp = jsonify(rst)
            resp.headers.extend(_cors_headers())
            return resp
        elif isinstance(rst, bytes):
            mimetype = 'application/octet-stream'
            if rst.startswith(b'\x89PNG'):
                mimetype = 'image/png'
            elif rst.startswith(b'\xff\xd8') and rst.endswith(b'\xff\xd9'):
                mimetype = 'image/jpeg'
            return Response(rst, mimetype=mimetype, headers=_cors_headers())
        elif callable(rst):
            return rst
        else:
            log.error('predict 函数返回了无效的结果：%s', rst)
            log.debug('有效的结果有 dict、list、tuple 对象，会被编码成 JSON；bytes、str 对象，会直接返回给客户端。')
            return {}, 500, _cors_headers()
    except ServingError as e:
        response = jsonify(e.payload)
        response.headers.extend(_cors_headers())
        response.status_code = e.status_code
        return response
    except HTTPException:
        raise
    except Exception:
        state = current_app.config.get('OPENBAYES_SERVING_DEBUGGER_STATE')
        log.exception('predict 函数发生了异常，请检查代码')
        if state:
            rv = state.collect_exception()
            log.error(f'使用浏览器打开 {rv["url"]} 调试此次失败的请求')
            return {'debug_url': rv['url']}, 500, _cors_headers()

        return {}, 500, _cors_headers()


@predict_blueprint.app_errorhandler(400)
def handle_bad_request(e):
    e
    return {'error': '无效的请求，请检查一下你的请求格式'}, 400


class Configurator:

    def limit_concurrency(self, n):
        self
        log.info(f'最大并发数限制为 {n}，超过并发限制的请求会返回 503 Service Unavailable')
        current_app.config['OPENBAYES_SERVING_UVICORN_CONFIG']['limit_concurrency'] = n


def make_app(predictor_cls):
    env = detect_env()
    if env == 'production':
        init_logging(logging.INFO)
    else:
        init_logging(logging.DEBUG)

    parser = argparse.ArgumentParser('openbayes-serv')
    parser.add_argument('--host', default='0.0.0.0', help='服务监听地址')

    port = 25252
    if env == 'gear':
        port = 8080
    elif env == 'production':
        port = 80

    parser.add_argument('--port', type=int, default=port, help='服务监听端口')

    app = Flask('model-serving')

    log.info('Openbayes Serving 正在启动...')

    with app.app_context():
        app.config['OPENBAYES_SERVING_UVICORN_CONFIG'] = {}

        lst = inspect.signature(predictor_cls).parameters.keys()
        clsname = getattr(predictor_cls, '__name__', repr(predictor_cls))
        args = {}
        for arg in lst:
            if arg not in INIT_ARG_HANDLERS:
                log.error(f"{clsname} 初始化失败: 不知道该如何提供 `{arg}`。支持的有：{list(INIT_ARG_HANDLERS.keys())}。请参考文档。")
                return

        for arg in lst:
            INIT_ARG_HANDLERS[arg].register_parameters(parser)

        options = parser.parse_args()
        app.config['OPENBAYES_SERVING_RUN_OPTIONS'] = options

        for arg in lst:
            args[arg] = INIT_ARG_HANDLERS[arg].init()

        user_obj = predictor_cls(**args)
        user_predict = getattr(user_obj, 'predict', None)
        if not user_predict:
            log.error(f"{clsname} 初始化失败: 没有实现 `predict` 接口。请参考文档。")
            return

        lst = inspect.signature(user_predict).parameters.keys()
        for arg in lst:
            if arg not in PREDICT_ARG_HANDLERS:
                log.error(f"{clsname} 执行失败: 在 `predict` 函数中，不知道该如何提供 `{arg}`。支持的有：{list(PREDICT_ARG_HANDLERS.keys())}。请参考文档。")
                return

        app.config['OPENBAYES_SERVING_PREDICTOR_OBJECT'] = user_obj
        app.config['OPENBAYES_SERVING_PREDICT_ARGS'] = lst

    if env == 'local':
        log.info('检测到本地开发环境，开启调试模式')
        debug.install(app)
    elif env == 'gear':
        log.info('检测到 Openbayes 算力容器环境，开启调试模式')
        prefix = None

        try:
            meta = requests.get('http://localhost:21999/gear-status').json()
            url = meta['links']['auxiliary']
            prefix = urlparse(url).path
            app.config['OPENBAYES_EXTERNAL_ROOT'] = url
            log.info('外部可访问的 URL：%s', url)
        except Exception:
            log.exception('无法获取 Openbayes 算力容器的元信息，不再进行对接，将不能正常从容器外部访问。')

        app.register_blueprint(predict_blueprint, url_prefix=prefix)
        debug.install(app, url_prefix=prefix)
    elif env == 'production':
        log.info('检测到生产环境，关闭调试模式。请求的统计信息可以在 Openbayes 的控制台上查看。')

    app.register_blueprint(predict_blueprint, url_prefix='/')

    return app


def run(predictor_cls):
    app = make_app(predictor_cls)
    options = app.config['OPENBAYES_SERVING_RUN_OPTIONS']

    config = uvicorn.Config(app,
        host=options.host, port=options.port,
        log_level='info', interface='wsgi',
        **app.config['OPENBAYES_SERVING_UVICORN_CONFIG']
    )

    logger = logging.getLogger('uvicorn')
    logger.handlers[:] = []

    patch_log_formatter(logging.getLogger('uvicorn.access'))
    logging.getLogger('uvicorn.error').setLevel(logging.ERROR)

    server = uvicorn.Server(config=config)
    log.info(f'Openbayes Serving 开始服务，本地访问地址：http://{options.host}:{options.port}')
    log.info('-' * 60)
    server.run()

# def obt.serv.check(pattern, obj=<payload>):
#     '''检查指定的对象是否满足 pattern，在 predict 处使用，默认检查 payload。要实现一个后门，用来给前端暴露这个 pattern 具体是什么。'''



# -*- coding: utf-8 -*-

# -- stdlib --
import functools
import os
import sys

# -- third party --
# -- own --

# -- code --
@functools.lru_cache(1)
def detect_env():
    if sys.platform == 'win32':
        return 'local'

    if os.environ.get('OPENBAYES_JOB_URL'):
        return 'gear'

    if os.environ.get('OPENBAYES_SERVING_PRODUCTION'):
        return 'production'

    return 'local'
